import typing as t
import typing_extensions as tex

from xpring.types import Seed, PrivateKey, PublicKey, Signature


class SigningAlgorithm(tex.Protocol):
    """
    Once ed25519 was added, the team wanted a way to differentiate seeds
    used with different signing algorithms. The seeds for both secp256k1 and
    ed25519 are arrays of 16 bytes, and thus could be encoded exactly the same
    way. It was decided to use a different prefix for ed25519, which yields
    encodings that always start with "sEd".
    """
    SEED_PREFIX: bytes

    def derive_key_pair(self, seed: Seed) -> t.Tuple[PrivateKey, PublicKey]:
        ...

    def sign(self, message: bytes, private_key: PrivateKey) -> Signature:
        ...

    def verify(
        self, message: bytes, signature: Signature, public_key: PublicKey
    ) -> bool:
        ...



def to_bytes(i: int, length: int) -> bytes:
    return i.to_bytes(length, byteorder='big', signed=False)


def from_bytes(bites: bytes) -> int:
    return int.from_bytes(bites, byteorder='big', signed=False)



from dataclasses import dataclass

import grpc
from xpring.proto.v1.get_account_info_pb2 import (
    GetAccountInfoRequest,
    GetAccountInfoResponse,
)
from xpring.proto.v1.account_pb2 import AccountAddress
from xpring.proto.v1.common_pb2 import Account
from xpring.proto.v1.get_fee_pb2 import (
    GetFeeRequest,
    GetFeeResponse,
)
from xpring.proto.v1.submit_pb2 import (
    SubmitTransactionRequest,
    SubmitTransactionResponse,
)
from xpring.proto.v1.xrp_ledger_pb2_grpc import XRPLedgerAPIServiceStub
from xpring.serialization import serialize_transaction
from xpring.types import Address, SignedTransaction


class Client:

    def __init__(self, grpc_client: XRPLedgerAPIServiceStub):
        self.grpc_client = grpc_client

    @classmethod
    def from_url(cls, grpc_url: str = 'grpc.xpring.tech:80'):
        channel = grpc.insecure_channel(grpc_url)
        grpc_client = XRPLedgerAPIServiceStub(channel)
        return cls(grpc_client)

    def get_account(self, address: Address) -> GetAccountInfoResponse:
        request = GetAccountInfoRequest(account=AccountAddress(address=address))
        return self.grpc_client.GetAccountInfo(request)

    def get_fee(self) -> GetFeeResponse:
        request = GetFeeRequest()
        return self.grpc_client.GetFee(request)

    def submit(
        self, signed_transaction: SignedTransaction
    ) -> SubmitTransactionResponse:
        blob = serialize_transaction(signed_transaction)
        request = SubmitTransactionRequest(signed_transaction=blob)
        return self.grpc_client.SubmitTransaction(request)



import typing as t

from fastecdsa import curve, ecdsa, keys
from fastecdsa.encoding.der import DEREncoder
# We use ecdsa only for verifying, until we can decode DER with fastecdsa:
# https://github.com/AntonKueltz/fastecdsa/issues/47
from ecdsa import curves, VerifyingKey
from ecdsa.util import sigdecode_der

from xpring import hashes
from xpring.algorithms.signing import Seed, PrivateKey, PublicKey, Signature
from xpring.bits import from_bytes, to_bytes

SEED_PREFIX = b'\x21'
FAMILY = bytes(4)

GROUP_ORDER = from_bytes(
    bytes.
    fromhex('FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141')
)
assert curve.secp256k1.q == GROUP_ORDER


def derive_private_key(seed: bytes) -> int:
    sequence = 0
    while True:
        buffer = seed + to_bytes(sequence, 4)
        private_key = from_bytes(hashes.sha512half(buffer))
        if private_key != 0 and private_key < GROUP_ORDER:
            return private_key
        sequence += 1


def compress_ecdsa_point(point) -> PublicKey:
    prefix = b'\x03' if point.y % 2 else b'\x02'
    return t.cast(PublicKey, prefix + to_bytes(point.x, 32))


def derive_key_pair(seed: Seed) -> t.Tuple[PrivateKey, PublicKey]:
    root_private_key = derive_private_key(seed)
    root_public_point = keys.get_public_key(root_private_key, curve.secp256k1)
    root_public_key = compress_ecdsa_point(root_public_point)

    inter_private_key = derive_private_key(root_public_key + FAMILY)
    inter_public_point = keys.get_public_key(inter_private_key, curve.secp256k1)
    inter_public_key = compress_ecdsa_point(inter_public_point)

    master_private_key = to_bytes(
        (root_private_key + inter_private_key) % GROUP_ORDER, 32
    )
    master_public_point = root_public_point + inter_public_point
    master_public_key = compress_ecdsa_point(master_public_point)

    return (
        t.cast(PrivateKey, master_private_key),
        t.cast(PublicKey, master_public_key),
    )


def sign(message: bytes, private_key: PrivateKey) -> Signature:
    digest = hashes.sha512half(message)
    signing_key = int.from_bytes(private_key, byteorder='big')
    r, s = ecdsa.sign(
        digest.hex(),
        signing_key,
        curve=curve.secp256k1,
        prehashed=True,
    )
    # Both (r, s) and (r, -s mod G = G - s) are valid, canonical signatures.
    # (r, s) is fully canonical only when s <= G - s.
    s_inverse = GROUP_ORDER - s
    if s > s_inverse:
        s = s_inverse
    signature = DEREncoder.encode_signature(r, s)
    return t.cast(Signature, signature)


def verify(message: bytes, signature: Signature, public_key: PublicKey) -> bool:
    digest = hashes.sha512half(message)
    # We need `ecdsa` to decompress the public key.
    # Would be nice if fastecdsa could do this.
    # TODO: Implement it ourselves.
    # https://github.com/AntonKueltz/fastecdsa/issues/47
    verifying_key = VerifyingKey.from_string(public_key, curves.SECP256k1)
    return verifying_key.verify(
        signature,
        digest,
        hashfunc=hashes.IdentityHash,
        sigdecode=sigdecode_der,
    )



from dataclasses import dataclass
import typing as t

import typing_extensions as tex

# 16 bytes.
# The two key-derivation functions we have today both use 128-bit keys,
# but only by happy accident.
# We should not assume that seeds will always be 128 bits.
Seed = t.NewType('Seed', bytes)
# Base58 encoding (with prefix and checksum) of
# a seed and a choice of signing algorithm.
EncodedSeed = t.NewType('EncodedSeed', str)

# 32 bytes.
PrivateKey = t.NewType('PrivateKey', bytes)

# 33 bytes.
PublicKey = t.NewType('PublicKey', bytes)
# 20 bytes computed from a public key.
AccountId = t.NewType('AccountId', bytes)
# Base58 encoding (with prefix and checksum) of an account ID.
# https://xrpl.org/accounts.html#address-encoding
Address = t.NewType('Address', str)

Signature = t.NewType('Signature', bytes)

NonXrpAmount = tex.TypedDict(
    'NonXrpAmount', {
        'value': str,
        'currency': str,
        'issuer': str,
    }
)

XrpAmount = str

Amount = t.Union[XrpAmount, NonXrpAmount]

# TODO: TypedDict instead of Mapping.
Transaction = t.Mapping
SignedTransaction = t.Mapping
